#ifndef MSP_GL_SL_VALIDATE_H_
#define MSP_GL_SL_VALIDATE_H_

#include <string>
#include <vector>
#include "glsl_error.h"
#include "visitor.h"

namespace Msp {
namespace GL {
namespace SL {

class Validator: protected TraversingVisitor
{
protected:
	Stage *stage;
	Node *last_provoker;

	Validator();

	void diagnose(Node &, Node &, Diagnostic::Severity, const std::string &);
	void diagnose(Node &n, Diagnostic::Severity s, const std::string &m) { diagnose(n, n, s, m); }
	void error(Node &n, const std::string &m) { diagnose(n, Diagnostic::ERR, m); }
	void add_info(Node &, const std::string &);
};

class TypeValidator: private Validator
{
private:
	bool in_struct;

public:
	TypeValidator();

	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
};

class DeclarationValidator: private Validator
{
private:
	typedef std::map<std::string, Statement *> BlockDeclarationMap;

	std::map<Block *, BlockDeclarationMap> declarations;
	std::map<std::string, InterfaceBlock *> interface_blocks;
	std::map<std::string, FunctionDeclaration *> overloaded_functions;
	bool anonymous_block;

public:
	DeclarationValidator();

	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	void multiple_definition(const std::string &, Statement &, Statement &);
	Statement *find_definition(const std::string &);
	void check_definition(const std::string &, Statement &);
	void record_definition(const std::string &, Statement &);

	virtual void visit(TypeDeclaration &);
	virtual void visit(BasicTypeDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(ImageTypeDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

class ReferenceValidator: private Validator
{
public:
	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(VariableReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(InterfaceBlockReference &);
	virtual void visit(FunctionCall &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

class ExpressionValidator: private Validator
{
public:
	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	virtual void visit(VariableDeclaration &);
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
