#include <stdexcept>
#include <msp/strings/format.h>
#include "light.h"
#include "matrix.h"
#include "misc.h"
#include "programdata.h"

using namespace std;

namespace Msp {
namespace GL {

Light::Light():
	diffuse(1),
	specular(1),
	position(0, 0, 1, 0),
	spot_dir(0, 0, -1),
	spot_exp(0),
	spot_cutoff(Geometry::Angle<float>::straight())
{
	attenuation[0] = 1;
	attenuation[1] = 0;
	attenuation[2] = 0;
}

void Light::update_matrix()
{
	Vector3 up_dir;
	if(20*abs(direction.z)>abs(direction.x)+abs(direction.y))
		up_dir.y = 1;
	else
		up_dir.z = 1;
	Vector3 right_dir = normalize(cross(direction, up_dir));

	Vector4 columns[4];
	columns[0] = compose(right_dir, 0.0f);
	columns[1] = compose(cross(right_dir, direction), 0.0f);
	columns[2] = compose(-direction, 0.0f);
	columns[3] = position;
	matrix = Matrix::from_columns(columns);
}

void Light::set_diffuse(const Color &c)
{
	diffuse = c;
}

void Light::set_specular(const Color &c)
{
	specular = c;
}

void Light::set_matrix(const Matrix &m)
{
	Placeable::set_matrix(m);
	position = matrix.column(3);
	spot_dir = normalize(-matrix.column(2).slice<3>(0));
	direction = (position.w ? spot_dir : normalize(-position.slice<3>(0)));
	update_matrix();
}

void Light::set_position(const Vector4 &p)
{
	position = p;
	if(!position.w)
		direction = normalize(-position.slice<3>(0));
	update_matrix();
}

void Light::set_spot_direction(const Vector3 &d)
{
	spot_dir = normalize(d);
	if(position.w)
		direction = spot_dir;
	update_matrix();
}

void Light::set_spot_exponent(float e)
{
	if(e<0)
		throw invalid_argument("Light::set_spot_exponent");

	spot_exp = e;
}

void Light::set_spot_cutoff(const Geometry::Angle<float> &c)
{
	if(c<Geometry::Angle<float>::zero() || (c>Geometry::Angle<float>::right() && c!=Geometry::Angle<float>::straight()))
		throw invalid_argument("Light::set_spot_cutoff");

	spot_cutoff = c;
}

void Light::disable_spot_cutoff()
{
	set_spot_cutoff(Geometry::Angle<float>::straight());
}

void Light::set_attenuation(float c, float l, float q)
{
	attenuation[0] = c;
	attenuation[1] = l;
	attenuation[2] = q;
}

void Light::update_shader_data(ProgramData &shdata, const Matrix &view_matrix, unsigned i) const
{
	string base = format("light_sources[%d]", i);
	shdata.uniform(base+".position", view_matrix*position);
	shdata.uniform(base+".diffuse", diffuse);
	shdata.uniform(base+".specular", specular);
}


Light::Loader::Loader(Light &l):
	DataFile::ObjectLoader<Light>(l)
{
	add("attenuation", &Loader::attenuation);
	add("diffuse", &Loader::diffuse);
	add("position", &Loader::position);
	add("specular", &Loader::specular);
	add("spot_direction", &Loader::spot_direction);
	add("spot_exponent", &Loader::spot_exponent);
	add("spot_cutoff", &Loader::spot_cutoff);
}

void Light::Loader::attenuation(float c, float l, float q)
{
	obj.set_attenuation(c, l, q);
}

void Light::Loader::diffuse(float r, float g, float b)
{
	obj.set_diffuse(Color(r, g, b));
}

void Light::Loader::position(float x, float y, float z, float w)
{
	obj.set_position(Vector4(x, y, z, w));
}

void Light::Loader::specular(float r, float g, float b)
{
	obj.set_specular(Color(r, g, b));
}

void Light::Loader::spot_direction(float x, float y, float z)
{
	obj.set_spot_direction(Vector3(x, y, z));
}

void Light::Loader::spot_exponent(float e)
{
	obj.set_spot_exponent(e);
}

void Light::Loader::spot_cutoff(float c)
{
	obj.set_spot_cutoff(Geometry::Angle<float>::from_degrees(c));
}

} // namespace GL
} // namespace Msp
